from abc import ABC, abstractmethod


class Trigger(ABC):
    @abstractmethod
    def should_trigger(self) -> bool:
        """ Returns the value of this trigger. True if it has triggered.

        Returns:
            bool -- The result.
        """

    @abstractmethod
    def persist_trigger(self):
        """ Performs any actions required by this trigger when the associated build step has been run.
        """



ENDC = "\033[0m"
RED = "\033[31m"
GREEN = "\033[32m"
YELLOW = "\033[33m"


class StepPrinter:
    """ Manages printing progress steps to stdout.
    """

    def __init__(self):
        self.current_step = ""

    def start_step(self, step: str):
        """ Start a new step, printing 'step ...'
        """

        self.current_step = step
        print("{}{} ...{}".format(YELLOW, self.current_step, ENDC))

    def end_step(self):
        """ End the most recently started step, printing 'step ✓'.
        """

        print("{}{} {}{}".format(GREEN, self.current_step, u"\u2713", ENDC))
        self.current_step = ""

    def skip_step(self, step: str):
        """ Skip a step, printing 'Skipping step ...'
        """

        print("{}{} (already complete) {}{}".format(GREEN, step, u"\u2933", ENDC))

    def error(self, code):
        """ Print the most recently started step with an error code
        """

        print(
            "{}{} exited with error code {} {}{}".format(
                RED, self.current_step, code, u"\u2717", ENDC
            )
        )
        self.current_step = ""



from abc import ABC, abstractmethod
from typing import Optional

from envy.lib.docker_manager import ContainerManager


class SetupStep(ABC):
    """ A Build Step is a build step for an ENVy environment image.
        Build steps have a trigger, which tells the builder when they need to be run.

    See Also: trigger.py
    """

    def __init__(self, name: str, label: Optional[str], container: ContainerManager):
        self._name = name
        self._label = label or name
        self._container = container
        self._has_run = False

    @property
    def name(self):
        return self._name

    @property
    def label(self):
        return self._label

    @abstractmethod
    def run(self):
        """ Run this build step.
            Does NOT check the trigger! Run should_trigger() first if you care.
        """
        self._has_run = True

    @abstractmethod
    def should_trigger(self) -> bool:
        """ Returns true if a trigger is configured, and that trigger is triggered.

        Returns:
            bool -- The result
        """
        raise NotImplementedError()

    @abstractmethod
    def persist_trigger(self):
        pass

    def has_built(self) -> bool:
        """ Returns true if this build step has build within this ENVy run.

        Returns:
            bool -- The result
        """
        return self._has_run



from .trigger import Trigger


class TriggerGroup(Trigger):
    """ A group of triggers. Will trigger if any of its children have triggered.
    """

    def __init__(self, children: [Trigger]):
        self.children = children

    def should_trigger(self) -> bool:
        result = False
        for child in self.children:
            result = result or child.should_trigger()

        return result

    def persist_trigger(self):
        for child in self.children:
            child.persist_trigger()



import requests
from envy.lib.config.file import find_config_file


class ConfigExecFile:
    def __init__(self, filename, byt):
        self.filename = filename
        self.bytes = byt


class FileDownloadError(Exception):
    def __init__(self, requests_error):
        super(FileDownloadError, self).__init__()
        self.requests_error = requests_error


def resolve_files(file_objects):
    """ Turn file objects from the config into "real" objects with Byte strings.
        Support URL and path formats
        Args:
            file_objects (list<dict>): fileObjects from the config
        Returns:
            list<ConfigExecFile>: List of executable files to run in the image
        Raises:
            FileDownloadError: When a file fails to download for some reason. Contains the Requests error.
    """
    if not file_objects:
        return None
    project_root = find_config_file().parent
    returned_list = []
    for obj in file_objects:
        try:
            if "url" in obj:
                r = requests.get(obj["url"])
                returned_list.append(ConfigExecFile(obj["filename"], r.content))
            elif "path" in obj:
                file_path = "{}/{}".format(project_root, obj["path"])
                try:
                    fil = open(filePath, "rb")
                    returned_list.append(ConfigExecFile(obj["filename"], fil.read()))
                except:
                    raise Exception("Failed opening file at " + file_path)
        except requests.exceptions.RequestException as e:
            raise FileDownloadError(e)
    return returned_list



from pathlib import Path
import os

from envy.lib.config import ENVY_PROJECT_DIR

_STATE_DIRECTORY = ".envy"

ENVY_STATE_PATH = "{}/{}".format(ENVY_PROJECT_DIR, _STATE_DIRECTORY)


def create_directory_if_not_exists():
    """ Creates the state directory if not found
    """
    if not os.path.isdir(ENVY_STATE_PATH):
        os.mkdir(ENVY_STATE_PATH)
        Path("{}/DO_NOT_COMMIT".format(ENVY_STATE_PATH)).touch()



import docker
from requests.exceptions import ConnectionError as RequestsConnectionError


class ConnectionTester:
    def __init__(self, dockerClient):
        self.client = dockerClient
        self.result = False
        self.message = ""

        self.__test()

    def __test(self):
        try:
            if self.client.ping():
                self.result = True
            else:
                self.result = False
                self.message = "Docker daemon is unavailable"
        except docker.errors.APIError as e:
            if e.status_code == 502:
                self.message = "Docker daemon is unavailable (502). Is it starting up?"
            elif e.status_code == 403:
                self.message = "Connection denied (403). Do you have permission to connect to docker?"
            elif e.status_code == 401:
                self.message = "Connection denied (401). Do you have permission to connect to docker?"
            else:
                self.message = "HTTP error (e.status_code)."
        except RequestsConnectionError as e:
            self.message = "Connection refused. Is docker running?"
        except Exception as e:  # pylint: disable=W0703
            self.message = str(e)

    def ok(self) -> bool:
        return self.result

    def reason(self) -> str:
        return self.message

    def print_err(self):
        print(self.message)



from .trigger import Trigger


class TriggerAlways(Trigger):
    """ A trigger that always triggers.
    """

    def __init__(self):
        pass

    def should_trigger(self) -> bool:
        return True

    def persist_trigger(self):
        pass



from hashlib import md5
import os
from pathlib import Path
from docker.types import Mount
from docker import DockerClient
from docker.models.containers import Container
import dockerpty

from envy.lib.config import ENVY_PROJECT_DIR, ENVY_CONFIG


class ContainerNotFound(Exception):
    pass


class ContainerNotRunning(Exception):
    pass


class ContainerError(Exception):
    def __init__(self, code):
        super(ContainerError, self).__init__(code)
        self.code = code


class ContainerManager:
    ### Static Container Creation ###
    @staticmethod
    def __generate_container_name() -> str:
        return f"envy-{ENVY_PROJECT_DIR.name}-{md5(str(ENVY_PROJECT_DIR).encode()).hexdigest()}-container"

    @staticmethod
    def create(docker_client: DockerClient, image_id: str) -> "ContainerManager":
        """ Creates a container with the given image id

        Arguments:
            docker_client {DockerClient} -- A docker client
            image_id {str} -- the ID of the image to use

        Returns:
            ContainerManager -- a container manager for the created container
        """
        print("Creating ENVy container")

        container = docker_client.containers.create(
            image_id,
            "tail -f /dev/null",
            name=ContainerManager.__generate_container_name(),
            network_mode="host",
            mounts=[
                Mount(
                    ENVY_CONFIG.get_project_mount_path(),
                    str(ENVY_PROJECT_DIR),
                    type="bind",
                ),
                Mount("/var/run/docker.sock", "/var/run/docker.sock", type="bind"),
            ],
        )

        return ContainerManager(docker_client, container.id)

    ### Container Management ###
    def __init__(self, docker_client: DockerClient, container_id: str):
        """ Creates a container manager for the given container id

        Arguments:
            docker_client {DockerClient} -- A docker client
            container_id {str} -- the container ID

        Returns:
            ContainerManager -- a container manager for the container
        """
        self.docker_client = docker_client

        self.container_id = container_id

    def __find(self) -> Container:
        for container in self.docker_client.containers.list(all=True):
            if container.id == self.container_id:
                return container
        return None

    def is_running(self) -> bool:
        """ Determines if the container is running

        Raises:
            ContainerNotFound: The container was not found

        Returns:
            bool -- Result
        """
        container = self.__find()

        if not container:
            raise ContainerNotFound

        return bool("running" in container.status)

    def exec(self, command: str, as_user: bool = False, relpath: str = None):
        """ Executes the command in the container

        Arguments:
            command {str} -- The command to run. Usually /bin/bash <>
            relpath {str} -- Relative path to project root to enter before executing. Optional, defaults to none.
        Raises:
            ContainerNotFound: The container was not found
            ContainerNotRunning: The container was not running
        """
        if not self.is_running():
            raise ContainerNotRunning()

        cdto = ENVY_CONFIG.get_project_mount_path()
        if relpath is not None:
            cdto = Path(ENVY_CONFIG.get_project_mount_path(), relpath)

        command_inside_project = "/bin/bash -c 'cd {}; {}'".format(
            cdto, command.replace("'", "'\\''")
        )

        if as_user:
            groups = ",".join(str(x) for x in os.getgroups())
            userspec = str(os.getuid()) + ":" + str(os.getgid())
            command_inside_project = "/usr/sbin/chroot --groups={} --userspec={} / /bin/bash --noprofile -c 'cd {}; {}'".format(
                groups,
                userspec,
                ENVY_CONFIG.get_project_mount_path(),
                command.replace("'", "'\\''"),
            )

        exit_code = dockerpty.exec_command(
            self.docker_client, self.container_id, command_inside_project
        )

        if exit_code != 0:
            raise ContainerError(exit_code)

    def ensure_running(self):
        """ Ensures that the container is running

        Raises:
            ContainerNotFound: The container was not found
        """
        container = self.__find()

        if not container:
            raise ContainerNotFound()

        if "running" not in container.status:
            container.start()

    def ensure_stopped(self):
        """ Ensures that the container is not running

        Raises:
            ContainerNotFound: The container was not found
        """
        container = self.__find()

        if not container:
            raise ContainerNotFound()

        if "running" in container.status:
            container.kill()

    def destroy(self):
        """ Destroys the container
        """
        container = self.__find()

        if container is not None:
            container.remove()



import sys

from schema import Schema, SchemaError, Optional, And, Or, Use

_DEFAULT_ENVIRONMENT_BASE = {"image": "ubuntu:18.04", "package-manager": "apt"}

_DEFAULT_PROJECT_DIR = "/project"

_DEFAULT_ENVIRONMENT = {
    "base": _DEFAULT_ENVIRONMENT_BASE,
    "project-dir": _DEFAULT_PROJECT_DIR,
    "system-packages": [],
    "setup-steps": [],
}

_STEP_TYPES = ["script", "remote"]

_SIMPLE_TRIGGERS = ["always"]

_DEFAULT_TRIGGERS = {"system-packages": [], "files": [], "steps": []}


def __validate_project_dir(project_dir: str) -> bool:
    if not project_dir:
        print("Project directory cannot be empty")
        return False
    if project_dir[0] != "/":
        print("Project directory must have an absolute root path")
        return False
    if len(project_dir) == 1:
        print("Root is not a valid project directory")
        return False

    return True


def __validate_setup_step(step: {}) -> bool:
    """ Validates that the given build step contains the appropriate keys given its type

    Arguments:
        step {{}} -- The build step dictionary

    Returns:
        bool -- result
    """
    if step["type"] == "script":
        return "run" in step
    if step["type"] == "remote":
        return "url" in step or "path" in step
    return False


def __validate_environment_setup_steps(environment: {}) -> bool:
    """ Validates that the build steps in the given environment dictionary are valid.
        Verifies that their names are unique, and that they only depend on previously defined steps and system packages

    Arguments:
        environment {{}} -- The environment dictionary

    Returns:
        bool -- Result
    """

    valid_system_package_names = {
        dep["recipe"] for dep in environment["system-packages"]
    }

    seen_step_names = set()

    for step in environment["setup-steps"]:
        if step["name"] in seen_step_names:
            print("Step names must be unique")
            return False
        seen_step_names.add(step["name"])

        if isinstance(step["triggers"], list):
            for system_package_trigger in step["triggers"]["system-packages"]:
                if system_package_trigger not in valid_system_package_names:
                    print(
                        "system package triggers can only depend on valid system packages"
                    )
                    return False

            for step_trigger in step["triggers"]["steps"]:
                if step_trigger not in seen_step_names:
                    print("Step triggers can only depend on previously defined steps")
                    return False
                if step_trigger == step["name"]:
                    print("Step triggers cannot depend on themselves")
                    return False

    return True


_SCHEMA = Schema(
    {
        Optional("environment", default=_DEFAULT_ENVIRONMENT): And(
            {
                Optional("base", default=_DEFAULT_ENVIRONMENT_BASE): {
                    "image": str,
                    Optional("package-manager"): str,
                },
                Optional("project-dir", default=_DEFAULT_PROJECT_DIR): And(
                    str, __validate_project_dir
                ),
                Optional("system-packages", default=[]): [
                    {"recipe": str, Optional("version"): Or(str, int, float)}
                ],
                Optional("setup-steps", default=[]): [
                    And(
                        {
                            "name": str,
                            "type": And(
                                str, Use(str.lower), lambda t: t in _STEP_TYPES
                            ),
                            Optional("label", default=None): str,
                            Optional("triggers", default=_DEFAULT_TRIGGERS): Or(
                                And(
                                    str, Use(str.lower), lambda t: t in _SIMPLE_TRIGGERS
                                ),
                                {
                                    Optional("system-packages", default=[]): [str],
                                    Optional("files", default=[]): [str],
                                    Optional("steps", default=[]): [str],
                                },
                            ),
                            Optional("as_user", default=True): bool,
                            Or("run", "url", "path", only_one=True): Or([str], str),
                        },
                        __validate_setup_step,
                    )
                ],
            },
            __validate_environment_setup_steps,
        ),
        Optional("actions", default=[]): [
            {
                "name": str,
                "script": str,
                "help": str,
                Optional("disable_relpath", default=False): bool,
            }
        ],
        Optional("services", default={}): {Optional("compose-file"): str},
    }
)


def validate(raw_data: str) -> {}:
    try:
        return _SCHEMA.validate(raw_data)
    except SchemaError as e:
        sys.exit(e.code)



import sys
import os

from .file import find_config_file, find_relative_cwd
from .envy_config import EnvyConfig

ENVY_CONFIG_FILE_PATH = find_config_file()
if ENVY_CONFIG_FILE_PATH is None:
    sys.stderr.write("Envy config file not found.\n")
    sys.exit(1)

ENVY_CURRENT_RELATIVE_PATH = find_relative_cwd()

ENVY_PROJECT_DIR = ENVY_CONFIG_FILE_PATH.parent

ENVY_CONFIG = EnvyConfig(ENVY_CONFIG_FILE_PATH)



import hashlib
import os

from envy.lib.state import ENVY_STATE

from .trigger import Trigger

CHUNK_SIZE = 4096


class TriggerWatchfile(Trigger):
    """ Triggers whenever the given file changes.
        Stores a hash of the file's contents in the state to determine changes.
    """

    def __init__(self, watchfile: str):
        self.watchfile = watchfile
        self.name_hash = self.__compute_name_hash()
        self.file_hash = self.__compute_file_hash()

    def should_trigger(self) -> bool:
        return ENVY_STATE.get_watchfile_hash(self.name_hash) != self.file_hash

    def persist_trigger(self):
        ENVY_STATE.set_watchfile_hash(self.name_hash, self.file_hash)

    def __compute_file_hash(self):
        md5 = hashlib.md5()

        if os.path.isfile(self.watchfile):
            with open(self.watchfile, "rb") as f:
                for chunk in iter(lambda: f.read(CHUNK_SIZE), b""):
                    md5.update(chunk)

        return md5.hexdigest()

    def __compute_name_hash(self):
        return hashlib.md5(self.watchfile.encode()).hexdigest()



from abc import abstractmethod

from envy.lib.docker_manager import ContainerManager

from .setup_step import SetupStep


class PackageManagerStep(SetupStep):
    """ A Package Manager Step is a Build Step that repesents a step that handles
    system package dependencies for an ENVy environment image.
    """

    def __init__(self, container: ContainerManager, packages: [{}]):
        super(PackageManagerStep, self).__init__(
            "System Packages", "Installing System Packages", container
        )
        self._packages = packages

    @abstractmethod
    def updated_packages(self) -> [str]:
        """ Return a list of packages that were changed or freshly installed
        If called before the step is run, will return an empty list
        """
        if self.has_built():
            return self._packages
        return []



from abc import abstractmethod
from typing import Optional

from envy.lib.triggers import Trigger
from envy.lib.state import ENVY_STATE
from envy.lib.docker_manager import ContainerManager

from .setup_step import SetupStep


class AssignableTriggerStep(SetupStep):
    """ A Build Step that allows its trigger to be set
    """

    def __init__(self, name: str, label: Optional[str], container: ContainerManager):
        super(AssignableTriggerStep, self).__init__(name, label, container)
        self._trigger = None

    def set_trigger(self, trigger: Trigger):
        """ Sets the trigger for this build step

        Arguments:
            trigger {Trigger} -- The trigger
        """
        self._trigger = trigger

    def should_trigger(self) -> bool:
        """ Returns true if a trigger is configured, and that trigger is triggered.

        Returns:
            bool -- The result
        """
        if self.name not in ENVY_STATE.get_run_steps():
            return True

        if self._trigger:
            return self._trigger.should_trigger()

        return False

    def persist_trigger(self):
        ENVY_STATE.add_run_step(self.name)
        if self._trigger:
            self._trigger.persist_trigger()

    @abstractmethod
    def run(self):
        super().run()



import os
import pwd
import grp

from collections import OrderedDict

from envy.lib.docker_manager import ContainerManager
from envy.lib.io import StepPrinter
from envy.lib.config import ENVY_CONFIG

import envy.lib.triggers as triggers

from .script_setup_step import ScriptSetupStep
from .remote_setup_step import RemoteSetupStep
from .apt_package_manager_step import AptPackageManagerStep


class Builder:
    """ Runs triggered build steps on a container
    """

    def __init__(self, container: ContainerManager, printer: StepPrinter):
        self.container = container
        self.printer = printer
        self.steps = OrderedDict()
        self.system_package_step = None

    def build(self):
        # Create system packages step
        self.__create_system_packages_step()

        # Create initial-setup step
        self.__create_initial_setup_steps()

        # Create steps
        self.__create_steps()

        # Run triggered steps
        self.__run_triggered()

        # Persist triggers
        self.__persist_triggers()

    def __create_system_packages_step(self):
        self.system_package_step = AptPackageManagerStep(
            self.container, ENVY_CONFIG.get_system_packages()
        )
        self.steps[self.system_package_step.name] = self.system_package_step

    def __create_initial_setup_steps(self):
        # Set up this user's username and groups inside the container.
        uid = os.getuid()
        gid = os.getgid()
        groups = os.getgroups()
        uname = pwd.getpwuid(uid).pw_name

        group_info = [grp.getgrgid(group) for group in groups]
        group_creation = [
            "echo '{}:x:{}:{}' >> /etc/group".format(
                group.gr_name, str(group.gr_gid), uname if group.gr_gid != gid else ""
            )
            for group in group_info
        ]

        chmod_step = ScriptSetupStep(
            "ENVY_chmod_root",
            "Setting up home environment",
            self.container,
            [
                "chmod a+wrx /root",
                "chmod a+wrx /",
                "echo '{}:x:{}:{}::/root:/bin/bash' >> /etc/passwd".format(
                    uname, str(uid), str(gid)
                ),
            ]
            + group_creation,
            False,
        )
        self.steps[chmod_step.name] = chmod_step

    def __create_steps(self):
        for m in ENVY_CONFIG.get_setup_steps():
            # Create step
            name = m["name"]
            label = m["label"]
            if m["type"] == "script":
                step = ScriptSetupStep(
                    name, label, self.container, m["run"], m["as_user"]
                )
            elif m["type"] == "remote":
                step = RemoteSetupStep(name, label, self.container, m["url"])

            # Create and register triggers
            if m["triggers"] == "always":
                trigger = triggers.TriggerAlways()
            else:
                trigger_list = []
                for t in m["triggers"]["system-packages"]:
                    trigger_list.append(
                        triggers.TriggerSystemPackage(t, self.system_package_step)
                    )
                for t in m["triggers"]["files"]:
                    trigger_list.append(triggers.TriggerWatchfile(t))
                for t in m["triggers"]["steps"]:
                    trigger_list.append(triggers.TriggerStep(self.steps[t]))

                trigger = triggers.TriggerGroup(trigger_list)

            step.set_trigger(trigger)

            # Add step to dict
            self.steps[name] = step

    def __run_triggered(self):
        for step in self.steps.values():
            if step.should_trigger():
                self.printer.start_step(step.label)
                step.run()
                self.printer.end_step()
            else:
                self.printer.skip_step(step.label)

    def __persist_triggers(self):
        for step in self.steps.values():
            if step.has_built():
                step.persist_trigger()



import subprocess


class ComposeManager:
    """ Manages a docker-compose file. Currently used for running sidecar services.
    """

    def __init__(self, compose_file: str):
        self.compose_file = compose_file

    # TODO compose may not be installed
    # TODO I don't think we can do anything about this, but we *will* leave images dangling here. Probably just something we're ok with.

    def up(self):
        """ Creates and starts the docker-compose environment.
            Also pulls new images for each service if necessary.
        """

        # TODO detect that pull is necessary and print something, since it can take a while
        try:
            subprocess.run(
                ["docker-compose", "-f", self.compose_file, "pull"],
                check=True,
                capture_output=True,
            )
            subprocess.run(
                ["docker-compose", "-f", self.compose_file, "up", "-d"],
                check=True,
                capture_output=True,
            )
        except subprocess.CalledProcessError as e:
            print(
                "Failed to start sidecar services. Command returned with error code {}".format(
                    e.returncode
                )
            )
            print(e.stdout.decode())
            print(e.stderr.decode())

    def down(self):
        """ Stops the docker-compose environment. Does not delete any containers or volumes.
        """

        try:
            subprocess.run(
                ["docker-compose", "-f", self.compose_file, "stop"],
                check=True,
                capture_output=True,
            )
        except subprocess.CalledProcessError as e:
            print(
                "Failed to stop sidecar services. Command returned with error code {}".format(
                    e.returncode
                )
            )
            print(e.stdout.decode())
            print(e.stderr.decode())

    def nuke(self):
        """ Destroys the docker-compose environment. Removes all service containers and volumes.
        """

        try:
            subprocess.run(
                ["docker-compose", "-f", self.compose_file, "kill"],
                check=True,
                capture_output=True,
            )
            subprocess.run(
                ["docker-compose", "-f", self.compose_file, "rm", "-sf"],
                check=True,
                capture_output=True,
            )
        except subprocess.CalledProcessError as e:
            print(
                "Failed to nuke sidecar services. Command returned with error code {}".format(
                    e.returncode
                )
            )
            print(e.stdout.decode())
            print(e.stderr.decode())



from envy.lib.setup_step.package_manager_step import PackageManagerStep

from .trigger import Trigger


class TriggerSystemPackage(Trigger):
    """ Triggers whenever the given system package is reinstalled.
        NOTE: Currently we don't reinstall system packages piecemeal, so this triggers whenever any package changes
    """

    def __init__(self, recipe: str, package_manager_step: PackageManagerStep):
        self.recipe = recipe
        self.package_manager_step = package_manager_step

    def should_trigger(self) -> bool:
        if self.recipe in [
            package["recipe"]
            for package in self.package_manager_step.updated_packages()
        ]:
            return True
        return False

    def persist_trigger(self):
        pass



import docker

from envy.lib.config import ENVY_CONFIG
from envy.lib.state import ENVY_STATE

from .connection_tester import ConnectionTester
from .container_manager import ContainerManager


class ContainerExists(Exception):
    pass


class DockerManager:
    """ Manages the docker connection, and handles creation and acquisition of containers and images.
        Gets the docker client from env (docker.from_env()) automatically when created.
        You probably want to run connection_ok() to verify that docker is available, since we're not throwing in the constructor.
    """

    def __init__(self):
        self.docker_client = docker.from_env()

        self.tester = ConnectionTester(self.docker_client)

    ### DOCKER CONNECTION ###
    def connection_ok(self) -> bool:
        """ Returns the result of the docker connection test
        Returns:
            bool -- The result of the connection test
        """
        return self.tester.ok()

    def print_connection_err(self):
        """ Prints an error message for a failed docker connection.
        """
        self.tester.print_err()

    ### ENSURE ###
    def ensure_container(self) -> ContainerManager:
        """ Ensures that an ENVy container exists.
            If no valid container is found, creates a new container and writes it to the state.
            This ensures that a valid image exists as well.

        Returns:
            ContainerManager -- A container manager for the ensured container
        """
        container = self.get_container()

        if container:
            return container

        return self.create_container()

    ### CREATE ###
    def create_container(self) -> ContainerManager:
        """ Creates an ENVy container. Requires that the container does not already exist, and that the image already exists.

        Raises:
            ContainerExists: The container ID specified in ENVY_STATE already exists.

        Returns:
            ContainerManager -- A container manager for the newly created container
        """
        if self.get_container():
            raise ContainerExists()

        # TODO: This likely should be done smarter: give the user a message if we need to download, etc
        self.docker_client.images.pull(ENVY_CONFIG.get_base_image())

        container_manager = ContainerManager.create(
            self.docker_client, ENVY_CONFIG.get_base_image()
        )

        ENVY_STATE.set_container_id(container_manager.container_id)

        return container_manager

    ### GET ###
    def get_container(self) -> ContainerManager:
        """ Returns a container manager for the container specified in ENVY_STATE

        Returns:
            ContainerManager -- The container manager
        """
        container_id = ENVY_STATE.get_container_id()

        if container_id:
            return ContainerManager(self.docker_client, ENVY_STATE.get_container_id())
        return None

    ### NUKE ###
    def nuke(self):
        """ Stops and deletes the container found in ENVY_STATE
        """
        container = self.get_container()

        if container:
            container.ensure_stopped()
            container.destroy()

        ENVY_STATE.set_container_id("")



from .trigger import Trigger
from .trigger_always import TriggerAlways
from .trigger_group import TriggerGroup
from .trigger_step import TriggerStep
from .trigger_system_package import TriggerSystemPackage
from .trigger_watchfile import TriggerWatchfile



import json
import os
import shutil


class EnvyState:
    """ Manages ENVy's state through some files in the project's root directory
        Currently tracking:
            Environment hash: a hash of ENVy's environment config
            Container ID: the current ENVy environment container
    """

    def __init__(self, dir_path):
        self.directory = dir_path
        self.state = None

    def nuke(self):
        """ Removes the state directory
        """
        shutil.rmtree(self.directory)

    def get_container_id(self) -> str:
        return self.__get(["container", "dockerid"])

    def set_container_id(self, new_id: str):
        self.__set(["container", "dockerid"], new_id)

    def get_installed_packages(self):
        return self.__get(["packages"])

    def set_installed_packages(self, packages: [{}]):
        self.__set(["packages"], packages)

    def get_run_steps(self):
        return self.__get(["run_steps"]) or []

    def add_run_step(self, name):
        steps = self.get_run_steps()
        if name not in steps:
            steps.append(name)
            self.__set(["run_steps"], steps)

    def get_watchfile_hash(self, file: str) -> str:
        return self.__get(["watchfile", file])

    def set_watchfile_hash(self, file: str, new_hash: str):
        self.__set(["watchfile", file], new_hash)

    def __set(self, keys, value):
        state = self.__get_state()

        final_key = keys.pop()
        final_dict = state
        for key in keys:
            if not key in final_dict:
                final_dict[key] = {}
            final_dict = final_dict[key]

        final_dict[final_key] = value

        self.__set_state(state)

    def __get(self, keys):
        state = self.__get_state()

        final_key = keys.pop()
        final_dict = state
        for key in keys:
            if not key in final_dict:
                return None
            final_dict = final_dict[key]

        return final_dict.get(final_key, None)

    def __set_state(self, new_state):
        path = self.__get_state_path()

        with open(path, "w") as f:
            json.dump(new_state, f)

        self.state = new_state

    def __get_state(self) -> dict:
        if self.state is None:
            self.state = self.__load_state()

        return self.state

    def __load_state(self) -> dict:
        path = self.__get_state_path()

        if not os.path.isfile(path):
            return {}

        with open(path, "r") as f:
            return json.load(f)

    def __get_state_path(self):
        return "{}/state.json".format(self.directory)



from pathlib import Path
from typing import Optional

import yaml

from .schema import validate as validate_schema


class EnvyConfig:
    """ Reads the envyfile as specified in schema.py.
        Data is available in EnvyConfig.data, or with some accessors for important data.

    See Also: schema.py
    """

    def __init__(self, file_path: Path):
        self.file = file_path

        with file_path.open() as f:
            raw_data = yaml.safe_load(f)

        self.data = validate_schema(raw_data)

    def get_base_image(self) -> str:
        return self.data["environment"]["base"]["image"]

    def get_package_manager(self) -> str:
        config_manager = self.data["environment"]["base"]["package-manager"]

        if not config_manager:
            config_manager = self.__guess_package_manager()

        return config_manager

    def __guess_package_manager(self) -> str:
        return "apt"

    def get_project_mount_path(self) -> str:
        path = self.data["environment"]["project-dir"]

        if path[-1] == "/":
            return path[: len(path) - 1]

        return path

    def get_system_packages(self) -> [{}]:
        return self.data["environment"]["system-packages"]

    def get_setup_steps(self) -> [{}]:
        return self.data["environment"]["setup-steps"]

    def get_actions(self) -> [{}]:
        return self.data["actions"]

    def get_services_compose_path(self) -> Optional[str]:
        return self.data["services"].get("compose-file")



from pathlib import Path

_CONFIG_FILE_NAME = ".envyfile"


def find_config_file() -> Path:
    """ Finds the envyfile by traversing upwards from the cwd until a file matching _CONFIG_FILE_NAME is found

    Returns:
        [Path] -- Path to the envyfile
    """
    current_path = Path.cwd()
    while True:
        config_path = current_path / _CONFIG_FILE_NAME
        if config_path.exists() and config_path.is_file():
            return config_path
        if current_path == current_path.parent:
            break
        current_path = current_path.parent


def find_relative_cwd() -> Path:
    """ Finds the relative current working directory based on the root of the ENVy project.
        Does not return a Path as the result is *not* a useful path object.

    Returns:
        [Path] -- relative path to cwd from project root
    """
    cwd = Path.cwd()
    return cwd.relative_to(find_config_file().parent)



from envy.lib.setup_step.setup_step import SetupStep

from .trigger import Trigger


class TriggerStep(Trigger):
    """ Triggers whenever the given step rebuilds.
    """

    def __init__(self, step: SetupStep):
        self.step = step

    def should_trigger(self) -> bool:
        return self.step.has_built()

    def persist_trigger(self):
        pass



#!/usr/bin/python3

import argparse

from envy.lib.config import ENVY_CONFIG, ENVY_CURRENT_RELATIVE_PATH
from envy.lib.state import ENVY_STATE, create_directory_if_not_exists
from envy.lib.docker_manager import ComposeManager, ContainerError, DockerManager
from envy.lib.setup_step import Builder
from envy.lib.io import StepPrinter

STATUS_MSG_NO_CONTAINER = "ENVy has not been initialized for this project. Please run `envy up` to install the ENVy environment."
STATUS_MSG_CONTAINER_STOPPED = (
    "ENVy is not running. Run `envy up` to start the ENVy environment."
)
STATUS_MSG_CONTAINER_READY = "ENVy environment is ready!"


def up_command(_args: argparse.Namespace, _unknow_args: [str]):
    printer = StepPrinter()
    try:
        printer.start_step("Finding Docker Container")
        docker_manager = DockerManager()
        printer.end_step()

        if not docker_manager.connection_ok():
            docker_manager.print_connection_err()
            return

        if not ENVY_STATE.get_container_id():
            printer.start_step("Creating ENVy environment")
            docker_manager.create_container()
            create_directory_if_not_exists()
            printer.end_step()

        printer.start_step("Starting Container")
        container = docker_manager.ensure_container()
        container.ensure_running()
        printer.end_step()

        step_builder = Builder(container, printer)
        step_builder.build()

        compose_path = ENVY_CONFIG.get_services_compose_path()
        if compose_path:
            printer.start_step("Starting Sidecar Services")
            ComposeManager(compose_path).up()
            printer.end_step()
    except ContainerError as err:
        printer.error(err.code)


def shell_command(_args: argparse.Namespace, _unknown_args: [str]):
    try:
        docker_manager = DockerManager()

        if not docker_manager.connection_ok():
            docker_manager.print_connection_err()
            return

        container = docker_manager.get_container()

        if not container:
            print(STATUS_MSG_NO_CONTAINER)
            return

        if not container.is_running():
            print(STATUS_MSG_CONTAINER_STOPPED)
            return

        container.exec(
            "/bin/bash", as_user=True, relpath=str(ENVY_CURRENT_RELATIVE_PATH)
        )
    except ContainerError as err:
        print("Shell exited with error code {}".format(err.code))


def down_command(_args: argparse.Namespace, _unknown_args: [str]):
    printer = StepPrinter()
    try:
        printer.start_step("Connecting to Docker")
        docker_manager = DockerManager()
        printer.end_step()

        if not docker_manager.connection_ok():
            docker_manager.print_connection_err()
            return

        printer.start_step("Finding Docker Container")
        container = docker_manager.get_container()
        printer.end_step()

        if container:
            printer.start_step("Stopping Docker Container")
            container.ensure_stopped()
            printer.end_step()

        compose_path = ENVY_CONFIG.get_services_compose_path()
        if compose_path:
            printer.start_step("Stopping Sidecar Services")
            ComposeManager(compose_path).down()
            printer.end_step()
    except ContainerError as err:
        printer.error(err.code)


def nuke_command(_args: argparse.Namespace, _unknown_args: [str]):
    printer = StepPrinter()
    try:
        printer.start_step("Destroying ENVy Environment")
        docker_manager = DockerManager()

        if not docker_manager.connection_ok():
            docker_manager.print_connection_err()
            return

        docker_manager.nuke()
        ENVY_STATE.nuke()

        compose_path = ENVY_CONFIG.get_services_compose_path()
        if compose_path:
            ComposeManager(compose_path).nuke()

        printer.end_step()
    except ContainerError as err:
        printer.error(err.code)


def status_command(_args: argparse.Namespace, _unknown_args: [str]):
    try:
        docker_manager = DockerManager()

        if not docker_manager.connection_ok():
            docker_manager.print_connection_err()
            return

        container = docker_manager.get_container()

        if not container:
            print(STATUS_MSG_NO_CONTAINER)
        elif not container.is_running():
            print(STATUS_MSG_CONTAINER_STOPPED)
        else:
            print("ENVy environment is running!")
    except ContainerError as err:
        print("Status exited with error code {}".format(err.code))


def run_script(
    _args: argparse.Namespace,
    unknown_args: [str],
    script: str,
    disable_relpath: bool = False,
):
    try:
        docker_manager = DockerManager()

        if not docker_manager.connection_ok():
            docker_manager.print_connection_err()
            return

        container = docker_manager.get_container()

        if not container:
            print(STATUS_MSG_NO_CONTAINER)
        elif not container.is_running():
            print(STATUS_MSG_CONTAINER_STOPPED)
        else:
            command = "{} {}".format(script, " ".join(unknown_args))

            cdto = ENVY_CURRENT_RELATIVE_PATH
            if disable_relpath:
                cdto = None
            container.exec(command, True, cdto)
    except ContainerError as err:
        print("Script exited with error code {}".format(err.code))


def build_custom_command_parser(subparsers, name: str, info: {}):
    parser_custom = subparsers.add_parser(name, help=info.get("help"), add_help=False)
    parser_custom.set_defaults(
        func=lambda args, unknown_args: run_script(
            args, unknown_args, info["script"], info["disable_relpath"]
        )
    )


def get_parser(actions: [{}]) -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description="Use ENVy to set up and use your development environment"
    )
    subparsers = parser.add_subparsers(dest="subparser_name")
    # Create 'up' parser
    parser_up = subparsers.add_parser(
        "up", help="Create and start the development environment"
    )
    parser_up.set_defaults(func=up_command)

    # Create 'shell' parser
    parser_shell = subparsers.add_parser(
        "shell", help="Enter a shell in the development environment"
    )
    parser_shell.set_defaults(func=shell_command)

    # Create 'down' parser
    parser_down = subparsers.add_parser("down", help="Stop the development environment")
    parser_down.set_defaults(func=down_command)

    # Create 'nuke' parser
    parser_nuke = subparsers.add_parser(
        "nuke", help="Destroy all ENVy data for this project"
    )
    parser_nuke.set_defaults(func=nuke_command)

    # Create 'status' parser
    parser_status = subparsers.add_parser("status", help="Show ENVy status")
    parser_status.set_defaults(func=status_command)

    # Create parsers for arbitrary custom commands
    for action in actions:
        build_custom_command_parser(subparsers, action["name"], action)

    return parser


def main():
    parser = get_parser(ENVY_CONFIG.get_actions())
    args, unknown = parser.parse_known_args()
    if args.subparser_name:
        args.func(args, unknown)
    else:
        parser.print_help()


if __name__ == "__main__":
    main()



from typing import Optional

from envy.lib.docker_manager import ContainerManager

from .assignable_trigger_step import AssignableTriggerStep


class ScriptSetupStep(AssignableTriggerStep):
    def __init__(
        self,
        name: str,
        label: Optional[str],
        container: ContainerManager,
        scripts: [str],
        as_user: bool = True,
    ):
        super().__init__(name, label, container)
        self._scripts = scripts
        self._as_user = as_user

    def run(self):
        super().run()

        if isinstance(self._scripts, str):
            self._container.exec(self._scripts, self._as_user)
        else:
            for step in self._scripts:
                self._container.exec(step, self._as_user)



from envy.lib.docker_manager import ContainerManager

from .assignable_trigger_step import AssignableTriggerStep


class RemoteSetupStep(AssignableTriggerStep):
    def __init__(self, name: str, container: ContainerManager, url: str):
        super().__init__(name, container)
        self.url = url

    def run(self):
        super().run()

        print("WARNING: Remote Build Steps not implemented!")  # TODO



from envy.lib.state import ENVY_STATE

from .package_manager_step import PackageManagerStep


class AptPackageManagerStep(PackageManagerStep):
    """ An apt implementation for a Package Manager Step
    """

    def updated_packages(self) -> [str]:
        # TODO: consider changing the implementation to be aware of partial updates
        if self.has_built():
            return self._packages
        return []

    def should_trigger(self) -> bool:
        return ENVY_STATE.get_installed_packages() != self._packages

    def persist_trigger(self):
        ENVY_STATE.set_installed_packages(self._packages)

    def run(self):
        super().run()
        versioned_packages = [
            "{}={}".format(package["recipe"], package["version"])
            if "version" in package
            else package["recipe"]
            for package in self._packages
        ]

        apt_string = " ".join(
            [versioned_package for versioned_package in versioned_packages]
        )
        self._container.exec(
            'apt-get update && DEBIAN_FRONTEND=noninteractive apt-get -o Dpkg::Options::="--force-confdef" -o Dpkg::Options::="--force-confold" install -y {} && rm -rf /var/lib/apt/lists/*'.format(
                apt_string
            )
        )
